/*
 * U9Decode - a program to extract and decode the Ultima 9 music files
 * from music.flx
 *
 * Copyright (C) 2000 by Josef Drexler <jdrexler@julian.uwo.ca>
 *
 * The program is distributed according to the terms of the GNU General
 * Public License, for more information please see the file LICENSE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

	// ADPCM block size
#define BLOCKSIZEIN 30

	// Number of bytes in the output file, corresponding to one block
#define BLOCKSIZEOUT (4*(BLOCKSIZEIN-2))

	// typedefs for easier porting
typedef unsigned long int U32;	// 32bit unsigned
typedef signed long int S32;	// 32bit signed

typedef unsigned short int U16;	// 16bit unsigned
typedef signed short int S16;	// 16bit signed

	// the data blocks
typedef unsigned char blockin[BLOCKSIZEIN];
typedef S16 blockout[BLOCKSIZEOUT/2];

	// state information for one channel
typedef struct {
	S32 osa, osb;		// os = old sample; need two samples for each step
	S32 a, b;		// coefficients
	U32 shift;		// volume factor, in -3dB steps
	} state;

	// Only four of the seven MS ADPCM coefficients are
	// used by this algorithm
S32 coeffA[4] = { 0, 240, 460, 392 };
S32 coeffB[4] = { 0, 0, -208, -220 };

// Global variables
char *filename = "music.flx";

U32 volincrease = 0;	// amplification factor, in +3dB steps
S32 overbound;		// largest clipping occuring due to amplification
int slot = -1;		// which slot to decode; -1 for all
int onlyshow = 0;	// only show the slots, do not decode

char *helptext =
	"U9DECODE [music.flx-file] [slot number] [-a ampl] [-s] [-h]\n"
	"Copyright (C) 2000 by Josef Drexler <jdrexler@julian.uwo.ca>\n"
	"Distributed under the GPL, see file COPYING for details.\n"
	"\n"
	"If the file name is not given, \"music.flx\" is assumed.\n"
	"If the slot number is not given, all songs are decoded.\n"
	"-a ampl: Amplify the song by this many 3dB steps, can be negative.\n"
	"-s: Only show which slots are which songs, do not decode.\n"
	"-h: Show this help.\n"
	"\n"
	"This program does not yet work with mono files (i.e. the sfx files)\n"
	"\n"
	"Make sure you have enough hard disk space available.\n"
	"About 250MB are needed for decoding all of the files.\n";


inline void initstate(state *s)
{
  s->osa = 0;	// start from "0" samples (silence)
  s->osb = 0;
}

inline void newblock(state *s, S32 coeff, U32 shift)
{
  s->shift = shift;
  s->a = coeffA[coeff];
  s->b = coeffB[coeff];
}

// make sure sample is within the allowable range (not really necessary...)

inline void bounds(S32 *u)
{
  if (*u > 32767) {
	if (*u > overbound)
		overbound = *u;
	*u = 32767;
  } else if (*u < -32768) {
	if (-*u > overbound)
		overbound = -*u;
	*u = -32768;
  }
}

	// decode two samples for one channel
inline void newsample(state *s, S32 data1, S32 data2)
{
  S32 ua = s->osa, ub = s->osb;

	// two shifts (left/right) necessary because nibble is signed
	// (they are equivalent to a sign extend that way)
	// ua is intentionally overwritten!
  ua = ( ( (data1 << 28) >> (s->shift + 8) ) + ub * s->a + ua * s->b + 0x80) >> 8;

  bounds(&ua);

  ub = ( ( (data2 << 28) >> (s->shift + 8) ) + ua * s->a + ub * s->b + 0x80) >> 8;

  bounds(&ub);

	// store new samples
  s->osa = ua;
  s->osb = ub;
}

inline S16 scale(S32 sample)
{
  if (volincrease) {
	if (volincrease > 0)
		sample <<= volincrease;
	else
		sample >>= volincrease;

	bounds(&sample);
  }

  return sample;
}

void writewavheader(int fout, S32 len)
{
  S32 value;

	// write the .wav header
  lseek(fout, 0, SEEK_SET);
  write(fout, "RIFF", 4);
  value = len - 8;
  write(fout, &value, 4);
  write(fout, "WAVEfmt ", 8);
  value = 16;	// fmt chunk size
  write(fout, &value, 4);
  value = 1;	// Format PCM
  write(fout, &value, 2);
  value = 2;	// Channels
  write(fout, &value, 2);
  value = 22050;	// Freq.
  write(fout, &value, 4);
  value = 88100;	// bytes per sec
  write(fout, &value, 4);
  value = 4;	// block align
  write(fout, &value, 2);
  value = 16;	// bits per sample
  write(fout, &value, 2);
  write(fout, "data", 4);
  value = len - 44;	// data chunk size
  write(fout, &value, 4);
}

int decodeslot(int fin, int slot)
{
  int indexin, indexout;
  S32 curblock, numblocks, offset, filelen;
  char filename[64];

  int fout;

  state L, R;	// Left and Right channel states

  blockin in;
  blockout out;

  overbound = 0;	// indicate no clipping has happened for this song yet

	// read data offset and length
  lseek(fin, 0x80 + slot * 8, SEEK_SET);
  if (read(fin, &offset, 4) != 4)
	offset = 0;
  if (read(fin, &filelen, 4) != 4)
	filelen = 0;

  if ( (offset == 0) || (filelen == 0) )
	return 1;	// empty slot

	// read the filename
  lseek(fin, offset + 4, SEEK_SET);
  read(fin, filename, 56);
  strcat(filename, ".wav");

  if (onlyshow) {
	printf("Slot %d: %s, size %ld\n", slot, filename, filelen);
	return 0;
  }

  printf("Decoding %d - %s: 000%%", slot, filename);

	// calculate total blocks
  curblock = 0;
  numblocks = (filelen - 56) / BLOCKSIZEIN;

	// create output file in BINARY mode
  _fmode = O_BINARY;
  fout = creat(filename, S_IWRITE);
  if (fout == -1) {
	perror(filename);
	return 0;
  }

	// make room for the .wav header
  if (write(fout, out, 44) != 44) {
	perror("\b\b\b\bWrite");
	close(fout);
	return 0;
  }

  initstate(&L);
  initstate(&R);

  for (curblock=0; curblock<numblocks; curblock++) {

		// give status report every 1024 blocks, in percent
	if (curblock % 1024 == 0)
		printf("\b\b\b\b%3d%%", curblock * 100 / numblocks);

	if (read(fin, in, BLOCKSIZEIN) != BLOCKSIZEIN) {
		perror("\b\b\b\bRead");
		return 0;
	}

		// read index and volume values
	newblock(&L, in[0] >> 4, in[1] >> 4);
	newblock(&R, in[0] & 0xf, in[1] & 0xf);

		// decode the 28 bytes of ADPCM data
	for (indexin=2, indexout=0; indexin<BLOCKSIZEIN; indexin+=2) {
		newsample(&L, in[indexin] >> 4, in[indexin+1] >> 4);
		newsample(&R, in[indexin] & 0xf, in[indexin+1] & 0xf);

		out[indexout++] = scale(L.osa);
		out[indexout++] = scale(R.osa);
		out[indexout++] = scale(L.osb);
		out[indexout++] = scale(R.osb);
	}

	if (write(fout, out, BLOCKSIZEOUT) != BLOCKSIZEOUT) {
		perror("\b\b\b\bWrite");
		close(fout);
		return 0;
	}
  }

  writewavheader(fout, tell(fout));

  close(fout);

  printf("\b\b\b\bDone.\n");

  if (overbound)
	printf("Warning: Amplification has caused %ld%% clipping. "
		"Reduce the -a factor.\n", overbound * 100 / 32768);

  return 0;
}

	// process the command line
void commandline(int argc, char **argv)
{
  int i;

  for (i=1; i<argc; i++) {
	if ( (argv[i][0] == '/') || (argv[i][0] == '-') ) {
		switch (tolower(argv[i][1])) {
			case 'a' :
				if ( (i+1 < argc) && (isdigit(argv[i+1][0])) )
					volincrease = atoi(argv[++i]);
				else {
					printf("Error: No amplification factor given!\n");
					exit(1);
				}
				break;
			case 's' :
				onlyshow = 1;
				break;
			default :
				printf(helptext);
				exit(1);
		}
	} else if (isdigit(argv[i][0]))
		slot = atoi(argv[i]);
	else
		filename = argv[i];
  }

}

int main(int argc, char **argv)
{
  int fin, slotfrom, slotto, slotmax;

  if ( (sizeof(S32) != 4) || (sizeof(S16) != 2) ) {
	printf("Compilation error, wrong data size! Adjust S32 etc.\n");
	return 1;
  }

  commandline(argc, argv);

  slotfrom = slotto = slot;

  fin = open(filename, O_RDONLY | O_BINARY);
  if (fin == -1) {
	perror(filename);
	return 1;
  }

  lseek(fin, 0x50, SEEK_SET);
  read(fin, &slotmax, sizeof(slotmax));

  if (slotfrom == -1)
	slotfrom = 0;
  if (slotto == -1)
	slotto = slotmax - 1;

  if ( (slotfrom < 0) || (slotto >= slotmax) ) {
	printf("Invalid slot number given. Allowable range is %d to %d.\n",
		0, slotmax - 1);
	return 1;
  }

  for (slot=slotfrom; slot<=slotto; slot++)
	if (decodeslot(fin, slot))
		if (slotfrom == slotto)	// only show it if decoding single file
			printf("Slot is empty!\n");

  close(fin);

  return 0;
}
